from __future__ import annotations

from pydantic import BaseModel
from tinydb import Query, TinyDB
from tinydb.table import Table

from compaipair.utils import get_db


class CompletionTemplate(BaseModel):
    name: str
    priming: str = None
    decorator: str = None
    _db: TinyDB
    _table: Table

    def __init__(self, **data):
        super().__init__(**data)
        self._db = get_db()
        self._table = self._db.table("templates")

    @staticmethod
    def find_template(name: str) -> CompletionTemplate | None:
        db = get_db()
        condition = Query().name == name
        templates_table = db.table("templates")
        result = templates_table.get(condition)
        if isinstance(result, list):
            result = result[0]
        return CompletionTemplate.model_validate(result)

    @staticmethod
    def find_templates(name: str | None) -> list[CompletionTemplate]:
        db = get_db()
        templates_table = db.table("templates")
        if name is not None:
            condition = Query().name == name
            results = templates_table.search(condition)
        else:
            results = templates_table.all()

        return [CompletionTemplate.model_validate(template) for template in results]

    def save(self) -> list[int]:
        """
        Persists the model object to the database. If the model object does not
        yet have an id, one is assigned before persisting.

        Returns:
            int: the id of the persisted object
        """
        query = Query()
        return self._table.upsert(
            self.model_dump(exclude_none=True), query.name == self.name
        )

    def update_template(self, **kwargs) -> list[int]:
        update_obj = {}

        for key, val in kwargs.items():
            if key in self.__dict__.keys() and val is not None:
                update_obj[key] = val

        self.__dict__.update(update_obj)
        return self.save()

    def prompt(self, question: str):
        return f"{self.priming}\n{question}\n{self.decorator}"



import click

from compaipair.config.cli_functions import (
    save_api_key,
    print_cache_location,
    show_api_key,
    init,
)


@click.group(name="config")
def config_api():
    pass


@config_api.command(name="save-api-key")
@click.argument("api_key")
def save_api_key_cli(api_key):
    save_api_key(api_key)


@config_api.command(name="show-api-key")
def show_api_key_cli():
    show_api_key()


@config_api.command(name="show-cache-location")
def print_cache_location_cli():
    print_cache_location()


@config_api.command(name="init")
@click.option("--api-key", help="Google MakerSuite's API Key")
def init_cli(api_key: str):
    init(api_key=api_key)



from compaipair.cli import compai
from compaipair.complete.commands import (
    complete_cli,
    available_models_cli,
    complete_template_cli,
)
from compaipair.config.commands import config_api
from compaipair.templates.commands import templates_api

# Core commands
compai.add_command(complete_cli)
compai.add_command(available_models_cli)
compai.add_command(complete_template_cli)

# Config commands
compai.add_command(config_api)

# Templates commands
compai.add_command(templates_api)



import os

import google.generativeai as palm
from google.api_core import retry
from tinydb import TinyDB

from compaipair.types.exceptions import NoApiKeyException

prompt_template = """
{priming}
{question}
{decorator}
"""


def db_path():
    return os.path.join(get_cache_path(), "db.json")


def get_db() -> TinyDB:
    if not os.path.exists(db_path()):
        with open(db_path(), "w"):
            pass
    if get_db.db is None:
        get_db.db = TinyDB(db_path())
    return get_db.db


get_db.db = None


def close_db():
    if get_db.db is not None:
        get_db().close()
    get_db.db = None


def get_cache_path():
    return (
        os.environ["COMPAI_CACHE_PATH"]
        if "COMPAI_CACHE_PATH" in os.environ
        else os.path.join(os.path.expanduser("~"), ".compai")
    )


def create_cache_dir():
    cache_path = get_cache_path()
    if not os.path.exists(cache_path):
        os.mkdir(cache_path)


def get_available_models():
    return palm.list_models()


def api_key_path():
    return os.path.join(get_cache_path(), "api_key")


def get_api_key():
    # Try to load api key from environment
    if "GOOGLE_GENERATIVEAI_API_KEY" in os.environ:
        return os.environ["GOOGLE_GENERATIVEAI_API_KEY"]
    try:
        # Load API key
        with open(api_key_path(), "r") as f:
            return f.read()
    except FileNotFoundError:
        raise NoApiKeyException


def try_read_file(potential_path):
    if potential_path is None:
        return None
    if os.path.exists(potential_path):
        with open(potential_path, "r") as f:
            return f.read()
    return None


def get_prompt(priming: str, question: str, decorator: str):
    priming_file_contents = try_read_file(priming)
    if priming_file_contents is not None:
        priming = priming_file_contents

    question_file_contents = try_read_file(question)
    if question_file_contents is not None:
        question = question_file_contents

    decorator_file_contents = try_read_file(decorator)
    if decorator_file_contents is not None:
        decorator = decorator_file_contents

    return prompt_template.format(
        priming=priming, question=question, decorator=decorator
    )


def prompt_code(priming: str, code_file: str, decorator: str):
    with open(code_file, "r") as f:
        code = f.read()
    return prompt_template.format(priming=priming, question=code, decorator=decorator)


def configure_palm_api(api_key: str = None):
    if api_key is None:
        api_key = get_api_key()

    palm.configure(api_key=api_key, transport="rest")


def print_completion_results(completion, prompt=None):
    if prompt:
        print("Completion generated for prompt: \n")
        print(prompt)
        print("########################################\n\n")
    print("Completion: \n")
    print(completion.result)


@retry.Retry()
def generate_text(prompt, model, temperature=0.0):
    return palm.generate_text(prompt=prompt, model=model, temperature=temperature)



import os

from compaipair.utils import get_cache_path, get_api_key, create_cache_dir


def print_cache_location():
    print(get_cache_path())


def api_key_path():
    return os.path.join(get_cache_path(), "api_key")


def save_api_key(api_key: str):
    with open(api_key_path(), "w") as f:
        f.write(api_key)


def show_api_key():
    print(get_api_key())


def init(api_key):
    if not os.path.exists(get_cache_path()):
        create_cache_dir()
    save_api_key(api_key)



from rich.console import Console
from rich.markdown import Markdown

from compaipair.types.completion_template import CompletionTemplate
from compaipair.utils import close_db

console = Console()


def new_template(name: str, priming: str = "", decorator: str = ""):
    return CompletionTemplate(name=name, priming=priming, decorator=decorator).save()


def edit_template(name: str, priming: str = None, decorator: str = None):
    existing_template = CompletionTemplate.find_template(name)
    return existing_template.update_template(priming=priming, decorator=decorator)


template_output_template = """# {template_name}
## Priming
{priming}
## Decorator
{decorator}
## Example prompt:
{example_prompt}"""


def show_templates(name: str = None, verbose: bool = False):
    templates = CompletionTemplate.find_templates(name)

    if verbose:
        sample_question = "How to iterate through a list of strings?"
        output = "\n---\n".join(
            [
                template_output_template.format(
                    template_name=template.name,
                    priming=template.priming,
                    decorator=template.decorator,
                    example_prompt=template.prompt(sample_question),
                )
                for template in templates
            ]
        )
    else:
        output = "\n".join([f"- {template.name}" for template in templates])
    console.print(Markdown(output))
    close_db()
